@@ -19,6 +19,10 @@ module Bundler
       Gem::Requirement.new(req_str).satisfied_by?(version)
     end
 
+    def initialize
+      @replaced_methods = {}
+    end
+
     def version
       self.class.version
     end
@@ -132,6 +136,14 @@ module Bundler
       Gem.path
     end
 
+    def reset
+      Gem::Specification.reset
+    end
+
+    def post_reset_hooks
+      Gem.post_reset_hooks
+    end
+
     def gem_cache
       gem_path.map {|p| File.expand_path("cache", p) }
     end
@@ -212,7 +224,7 @@ module Bundler
       # Fetch all specs, minus prerelease specs
       spec_list = fetch_specs(true, false)
       # Then fetch the prerelease specs
-      fetch_prerelease_specs.each {|k, v| spec_list[k].push(*v) }
+      fetch_prerelease_specs.each {|k, v| spec_list[k].concat(v) }
 
       spec_list.values.first
     ensure
@@ -265,7 +277,9 @@ module Bundler
     def download_gem(spec, uri, path)
       uri = Bundler.settings.mirror_for(uri)
       fetcher = Gem::RemoteFetcher.new(configuration[:http_proxy])
-      fetcher.download(spec, uri, path)
+      Bundler::Retry.new("download gem #{uri}", Gem::RemoteFetcher::FetchError).attempts do
+        fetcher.download(spec, uri, path)
+      end
     end
 
     def security_policy_keys
@@ -283,13 +297,11 @@ module Bundler
 
     def reverse_rubygems_kernel_mixin
       # Disable rubygems' gem activation system
-      ::Kernel.class_eval do
-        if private_method_defined?(:gem_original_require)
-          alias_method :rubygems_require, :require
-          alias_method :require, :gem_original_require
+      kernel = (class << ::Kernel; self; end)
+      [kernel, ::Kernel].each do |k|
+        if k.private_method_defined?(:gem_original_require)
+          redefine_method(k, :require, k.instance_method(:gem_original_require))
         end
-
-        undef gem
       end
     end
 
@@ -298,41 +310,44 @@ module Bundler
 
       executables = specs.map(&:executables).flatten
 
-      ::Kernel.send(:define_method, :gem) do |dep, *reqs|
-        if executables.include? File.basename(caller.first.split(":").first)
-          break
-        end
-        reqs.pop if reqs.last.is_a?(Hash)
-
-        unless dep.respond_to?(:name) && dep.respond_to?(:requirement)
-          dep = Gem::Dependency.new(dep, reqs)
-        end
-
-        spec = specs.find {|s| s.name == dep.name }
-
-        if spec.nil?
-
-          e = Gem::LoadError.new "#{dep.name} is not part of the bundle. Add it to Gemfile."
-          e.name = dep.name
-          if e.respond_to?(:requirement=)
-            e.requirement = dep.requirement
-          else
-            e.version_requirement = dep.requirement
+      kernel = (class << ::Kernel; self; end)
+      [kernel, ::Kernel].each do |kernel_class|
+        redefine_method(kernel_class, :gem) do |dep, *reqs|
+          if executables.include? File.basename(caller.first.split(":").first)
+            break
           end
-          raise e
-        elsif dep !~ spec
-          e = Gem::LoadError.new "can't activate #{dep}, already activated #{spec.full_name}. " \
-                                 "Make sure all dependencies are added to Gemfile."
-          e.name = dep.name
-          if e.respond_to?(:requirement=)
-            e.requirement = dep.requirement
-          else
-            e.version_requirement = dep.requirement
-          end
-          raise e
-        end
+          reqs.pop if reqs.last.is_a?(Hash)
 
-        true
+          unless dep.respond_to?(:name) && dep.respond_to?(:requirement)
+            dep = Gem::Dependency.new(dep, reqs)
+          end
+
+          spec = specs.find {|s| s.name == dep.name }
+
+          if spec.nil?
+
+            e = Gem::LoadError.new "#{dep.name} is not part of the bundle. Add it to Gemfile."
+            e.name = dep.name
+            if e.respond_to?(:requirement=)
+              e.requirement = dep.requirement
+            else
+              e.version_requirement = dep.requirement
+            end
+            raise e
+          elsif dep !~ spec
+            e = Gem::LoadError.new "can't activate #{dep}, already activated #{spec.full_name}. " \
+                                   "Make sure all dependencies are added to Gemfile."
+            e.name = dep.name
+            if e.respond_to?(:requirement=)
+              e.requirement = dep.requirement
+            else
+              e.version_requirement = dep.requirement
+            end
+            raise e
+          end
+
+          true
+        end
       end
     end
 
@@ -362,19 +377,21 @@ module Bundler
     def replace_bin_path(specs)
       gem_class = (class << Gem; self; end)
 
-      redefine_method(gem_class, :find_spec_for_exe) do |name, *args|
+      redefine_method(gem_class, :find_spec_for_exe) do |gem_name, *args|
         exec_name = args.first
 
         spec = if exec_name
-          specs.find {|s| s.executables.include?(exec_name) }
+          specs.find {|s| s.name == gem_name && s.executables.include?(exec_name) } ||
+            specs.find {|s| s.executables.include?(exec_name) }
         else
-          specs.find {|s| s.name == name }
+          specs.find {|s| s.name == gem_name }
         end
         raise(Gem::Exception, "can't find executable #{exec_name}") unless spec
         raise Gem::Exception, "no default executable for #{spec.full_name}" unless exec_name ||= spec.default_executable
         unless spec.name == name
-          warn "Bundler is using a binstub that was created for a different gem.\n" \
-            "This is deprecated, in future versions you may need to `bundle binstub #{name}` " \
+          Bundler::SharedHelpers.major_deprecation \
+            "Bundler is using a binstub that was created for a different gem.\n" \
+            "You should run `bundle binstub #{gem_name}` " \
             "to work around a system/bundle conflict."
         end
         spec
@@ -461,11 +478,32 @@ module Bundler
       end
     end
 
-    def redefine_method(klass, method, &block)
-      if klass.instance_methods(false).include?(method)
-        klass.send(:remove_method, method)
+    def undo_replacements
+      @replaced_methods.each do |(sym, klass), method|
+        redefine_method(klass, sym, method)
+      end
+      post_reset_hooks.reject! do |proc|
+        proc.binding.eval("__FILE__") == __FILE__
+      end
+      @replaced_methods.clear
+    end
+
+    def redefine_method(klass, method, unbound_method = nil, &block)
+      begin
+        if (instance_method = klass.instance_method(method)) && method != :initialize
+          # doing this to ensure we also get private methods
+          klass.send(:remove_method, method)
+        end
+      rescue NameError
+        # method isn't defined
+        nil
+      end
+      @replaced_methods[[method, klass]] = instance_method
+      if unbound_method
+        klass.send(:define_method, method, unbound_method)
+      elsif block
+        klass.send(:define_method, method, &block)
       end
-      klass.send(:define_method, method, &block)
     end
 
     # Rubygems 1.4 through 1.6
@@ -481,11 +519,11 @@ module Bundler
       def stub_rubygems(specs)
         # Rubygems versions lower than 1.7 use SourceIndex#from_gems_in
         source_index_class = (class << Gem::SourceIndex; self; end)
-        source_index_class.send(:define_method, :from_gems_in) do |*args|
-          source_index = Gem::SourceIndex.new
-          source_index.spec_dirs = *args
-          source_index.add_specs(*specs)
-          source_index
+        redefine_method(source_index_class, :from_gems_in) do |*args|
+          Gem::SourceIndex.new.tap do |source_index|
+            source_index.spec_dirs = *args
+            source_index.add_specs(*specs)
+          end
         end
       end
 
@@ -502,6 +540,13 @@ module Bundler
         # which is too strict for the kinds of checks we care about. As a
         # result, validation is disabled on versions of RubyGems below 1.7.
       end
+
+      def post_reset_hooks
+        []
+      end
+
+      def reset
+      end
     end
 
     # Rubygems versions 1.3.6 and 1.3.7
@@ -605,7 +650,7 @@ module Bundler
         specs = fetch_specs(source, remote, "specs")
         pres = fetch_specs(source, remote, "prerelease_specs") || []
 
-        specs.push(*pres)
+        specs.concat(pres)
       end
 
       def download_gem(spec, uri, path)
@@ -680,25 +725,23 @@ module Bundler
     end
   end
 
-  if RubygemsIntegration.provides?(">= 2.1.0")
-    @rubygems = RubygemsIntegration::MoreFuture.new
-  elsif RubygemsIntegration.provides?(">= 1.99.99")
-    @rubygems = RubygemsIntegration::Future.new
-  elsif RubygemsIntegration.provides?(">= 1.8.20")
-    @rubygems = RubygemsIntegration::MoreModern.new
-  elsif RubygemsIntegration.provides?(">= 1.8.5")
-    @rubygems = RubygemsIntegration::Modern.new
-  elsif RubygemsIntegration.provides?(">= 1.8.0")
-    @rubygems = RubygemsIntegration::AlmostModern.new
-  elsif RubygemsIntegration.provides?(">= 1.7.0")
-    @rubygems = RubygemsIntegration::Transitional.new
-  elsif RubygemsIntegration.provides?(">= 1.4.0")
-    @rubygems = RubygemsIntegration::Legacy.new
-  else # Rubygems 1.3.6 and 1.3.7
-    @rubygems = RubygemsIntegration::Ancient.new
-  end
-
-  class << self
-    attr_reader :rubygems
+  def self.rubygems
+    @rubygems ||= if RubygemsIntegration.provides?(">= 2.1.0")
+      RubygemsIntegration::MoreFuture.new
+    elsif RubygemsIntegration.provides?(">= 1.99.99")
+      RubygemsIntegration::Future.new
+    elsif RubygemsIntegration.provides?(">= 1.8.20")
+      RubygemsIntegration::MoreModern.new
+    elsif RubygemsIntegration.provides?(">= 1.8.5")
+      RubygemsIntegration::Modern.new
+    elsif RubygemsIntegration.provides?(">= 1.8.0")
+      RubygemsIntegration::AlmostModern.new
+    elsif RubygemsIntegration.provides?(">= 1.7.0")
+      RubygemsIntegration::Transitional.new
+    elsif RubygemsIntegration.provides?(">= 1.4.0")
+      RubygemsIntegration::Legacy.new
+    else # Rubygems 1.3.6 and 1.3.7
+      RubygemsIntegration::Ancient.new
+    end
   end
 end
